from gemd.util.impl import recursive_flatmap
from gemd.entity.object import (
    ProcessSpec,
    ProcessRun,
    MaterialSpec,
    MeasurementSpec,
    MeasurementRun,
)
from citrine.resources.data_concepts import DataConcepts
from gemd.entity.value.base_value import BaseValue
from typing import Mapping, List
from gemd.entity.attribute import PropertyAndConditions


def make_attribute_table(gems: List[DataConcepts]) -> List[Mapping[str, BaseValue]]:
    """[ALPHA] the current status of make_attribute_table.

    Given a list of GEMD Objects, this method returns a list of dictionaries
    where in each dictionary there are keys of "object", "object_type" and keys for each
    unique Attribute Type + Name pair found within those Objects.
    Each value contains the associated Object, Object Type, or BaseValue.
    This dictionary can easily be converted into a Pandas DataFrame where there is a
    row for each Object, and the values within the cells are that Object's type or
    the various GEMD Value Types.

    This function will help users identify attribute usage across their Objects,
    and get a sense of the consistency of the data found in their GEMD objects.
    With the returned dictionary, one can easily do further analysis on both
    attribute values and value type consistency to assist in data cleaning and
    Attribute Template creation.

    Parameters
    ----------
    gems : List[DataConcepts]
        List of GEMD objects whose attributes you would like to compare.

    Returns
    -------
    List[Mapping[str, BaseValue]]
        A list of dictionaries where each dictionary represents an object and its attributes.

    """
    flattened_gems = recursive_flatmap(
        obj=gems, func=lambda x: [x], unidirectional=False
    )
    types_with_attributes = (
        ProcessSpec,
        ProcessRun,
        MaterialSpec,
        MeasurementSpec,
        MeasurementRun,
    )
    all_rows = []
    attributed_gems = [
        x for x in flattened_gems if isinstance(x, types_with_attributes)
    ]
    for gem in attributed_gems:
        row_dict = {"object": gem, "object_type": type(gem).__name__}
        if hasattr(gem, "conditions"):
            for cond in gem.conditions:
                row_dict[f"CONDITION: {cond.name}"] = cond.value
        if hasattr(gem, "parameters"):
            for param in gem.parameters:
                row_dict[f"PARAMETER: {param.name}"] = param.value
        if hasattr(gem, "properties"):
            for prop in gem.properties:
                if isinstance(prop, PropertyAndConditions):
                    row_dict[f"PROPERTY: {prop.property.name}"] = prop.property.value
                    for cond in prop.conditions:
                        row_dict[f"CONDITION: {cond.name}"] = cond.value
                else:
                    row_dict[f"PROPERTY: {prop.name}"] = prop.value
        all_rows.append(row_dict)
    return all_rows



from deprecation import deprecated
from functools import partial
from typing import Iterable, List, Optional
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._rest.pageable import Pageable
from citrine._rest.paginator import Paginator
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.design_candidate import DesignCandidate
from citrine.informatics.scores import Score


class DesignExecution(Resource['DesignExecution'], Pageable, AsynchronousObject):
    """The execution of a DesignWorkflow.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Design executions also have a ``status_description`` field with more information.

    """

    _paginator: Paginator = Paginator()
    _collection_key = 'response'
    _in_progress_statuses = ["INPROGRESS"]
    _succeeded_statuses = ["SUCCEEDED"]
    _failed_statuses = ["FAILED"]
    _session: Optional[Session] = None
    project_id: Optional[UUID] = None
    """:Optional[UUID]: Unique ID of the project that contains this workflow execution."""

    uid: UUID = properties.UUID('id', serializable=False)
    """:UUID: Unique identifier of the workflow execution"""
    workflow_id = properties.UUID('workflow_id', serializable=False)
    """:UUID: Unique identifier of the workflow that was executed"""
    version_number = properties.Integer("version_number", serializable=False)
    """:int: Integer identifier that increases each time the workflow is executed. The first
    execution has version_number = 1."""

    status = properties.Optional(properties.String(), 'status', serializable=False)
    """:Optional[str]: short description of the execution's status"""
    status_description = properties.Optional(
        properties.String(), 'status_description', serializable=False)
    """:Optional[str]: more detailed description of the execution's status"""
    status_info = properties.Optional(
        properties.List(properties.String()),
        'status_info',
        serializable=False
    )
    """:Optional[List[str]]: human-readable explanations of the status"""
    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    updated_by = properties.Optional(properties.UUID, 'updated_by', serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    create_time = properties.Optional(properties.Datetime, 'create_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""
    update_time = properties.Optional(properties.Datetime, 'update_time', serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    score = properties.Object(Score, 'score')
    """:Score: score by which this execution was evaluated"""
    descriptors = properties.List(properties.Object(Descriptor), 'descriptors')
    """:List[Descriptor]: all of the descriptors in the candidates generated by this execution"""

    def __init__(self):
        """Design executions are not directly instantiated by the user."""
        pass  # pragma: no cover

    def __str__(self):
        return '<DesignExecution {!r}>'.format(str(self.uid))

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/design-workflows/{workflow_id}/executions/{execution_id}',
            project_id=self.project_id,
            workflow_id=self.workflow_id,
            execution_id=self.uid
        )

    @classmethod
    def _build_candidates(cls, subset_collection: Iterable[dict]) -> Iterable[DesignCandidate]:
        for candidate in subset_collection:
            yield DesignCandidate.build(candidate)

    def candidates(self, *,
                   page: Optional[int] = None,
                   per_page: int = 100,
                   ) -> Iterable[DesignCandidate]:
        """Fetch the Design Candidates for the particular execution, paginated."""
        path = self._path() + '/candidates'

        fetcher = partial(self._fetch_page, path=path, fetch_func=self._session.get_resource)

        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_candidates,
                                        page=page,
                                        per_page=per_page)

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental(self) -> bool:
        """[DEPRECATED] whether the execution is experimental (newer, less well-tested)."""  # noqa - insisting this docstring is a signature
        return False

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental_reasons(self) -> List[str]:
        """[DEPRECATED] human-readable reasons why the execution is experimental."""
        return []



"""Resources that represent both individual and collections of projects."""
from functools import partial
from typing import Optional, Dict, List, Union, Iterable, Tuple, Iterator
from uuid import UUID
from warnings import warn

from gemd.entity.base_entity import BaseEntity
from gemd.entity.link_by_uid import LinkByUID

from citrine._rest.collection import Collection
from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url, use_teams, v1_deprecation_warn
from citrine.exceptions import NonRetryableException, ModuleRegistrationFailedException
from citrine.resources.api_error import ApiError
from citrine.resources.branch import BranchCollection
from citrine.resources.condition_template import ConditionTemplateCollection
from citrine.resources.dataset import DatasetCollection
from citrine.resources.delete import _async_gemd_batch_delete
from citrine.resources.descriptors import DescriptorMethods
from citrine.resources.design_space import DesignSpaceCollection
from citrine.resources.design_workflow import DesignWorkflowCollection
from citrine.resources.gemd_resource import GEMDResourceCollection
from citrine.resources.gemtables import GemTableCollection
from citrine.resources.ingredient_run import IngredientRunCollection
from citrine.resources.ingredient_spec import IngredientSpecCollection
from citrine.resources.material_run import MaterialRunCollection
from citrine.resources.material_spec import MaterialSpecCollection
from citrine.resources.material_template import MaterialTemplateCollection
from citrine.resources.measurement_run import MeasurementRunCollection
from citrine.resources.measurement_spec import MeasurementSpecCollection
from citrine.resources.measurement_template import MeasurementTemplateCollection
from citrine.resources.module import ModuleCollection
from citrine.resources.parameter_template import ParameterTemplateCollection
from citrine.resources.predictor import PredictorCollection
from citrine.resources.predictor_evaluation_execution import \
    PredictorEvaluationExecutionCollection
from citrine.resources.predictor_evaluation_workflow import \
    PredictorEvaluationWorkflowCollection
from citrine.resources.process_run import ProcessRunCollection
from citrine.resources.process_spec import ProcessSpecCollection
from citrine.resources.process_template import ProcessTemplateCollection
from citrine.resources.processor import ProcessorCollection
from citrine.resources.project_member import ProjectMember
from citrine.resources.project_roles import MEMBER, ROLES, ACTIONS
from citrine.resources.property_template import PropertyTemplateCollection
from citrine.resources.response import Response
from citrine.resources.table_config import TableConfigCollection
from citrine.resources.user import User


class Project(Resource['Project']):
    """
    A Citrine Project.

    A project is a collection of datasets and AI assets, some of which belong directly
    to the project, and some of which have been shared with the project.

    Parameters
    ----------
    name: str
        Name of the project.
    description: str
        Long-form description of the project.
    session: Session, optional
        The Citrine session used to connect to the database.

    Attributes
    ----------
    uid: UUID
        Unique uuid4 identifier of this project.
    status: str
        Status of the project.
    created_at: int
        Time the project was created, in seconds since epoch.

    """

    _response_key = 'project'
    _resource_type = ResourceTypeEnum.PROJECT

    name = properties.String('name')
    description = properties.Optional(properties.String(), 'description')
    uid = properties.Optional(properties.UUID(), 'id')
    status = properties.Optional(properties.String(), 'status')
    created_at = properties.Optional(properties.Datetime(), 'created_at')
    team_id = properties.Optional(properties.UUID, "team.id", serializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: Optional[str] = None,
                 session: Optional[Session] = None,
                 team_id: Optional[UUID] = None):
        self.name: str = name
        self.description: Optional[str] = description
        self.session: Session = session
        self.team_id: Optional[UUID] = team_id

    def _post_dump(self, data: dict) -> dict:
        return {key: value for key, value in data.items() if value is not None}

    def __str__(self):
        return '<Project {!r}>'.format(self.name)

    def _path(self):
        return format_escaped_url('/projects/{project_id}', project_id=self.uid)

    @property
    def modules(self) -> ModuleCollection:
        """Return a resource representing all visible design spaces."""
        return ModuleCollection(self.uid, self.session)

    @property
    def branches(self) -> BranchCollection:
        """Return a resource representing all visible branches."""
        return BranchCollection(self.uid, self.session)

    @property
    def design_spaces(self) -> DesignSpaceCollection:
        """Return a resource representing all visible design spaces."""
        return DesignSpaceCollection(self.uid, self.session)

    @property
    def processors(self) -> ProcessorCollection:
        """Return a resource representing all visible processors."""
        return ProcessorCollection(self.uid, self.session)

    @property
    def predictors(self) -> PredictorCollection:
        """Return a resource representing all visible predictors."""
        return PredictorCollection(self.uid, self.session)

    @property
    def descriptors(self) -> DescriptorMethods:
        """Return a resource containing a set of methods returning descriptors."""
        return DescriptorMethods(self.uid, self.session)

    @property
    def predictor_evaluation_workflows(self) -> PredictorEvaluationWorkflowCollection:
        """Return a collection representing all visible predictor evaluation workflows."""
        return PredictorEvaluationWorkflowCollection(self.uid, self.session)

    @property
    def predictor_evaluation_executions(self) -> PredictorEvaluationExecutionCollection:
        """Return a collection representing all visible predictor evaluation executions."""
        return PredictorEvaluationExecutionCollection(project_id=self.uid, session=self.session)

    @property
    def design_workflows(self) -> DesignWorkflowCollection:
        """Return a collection representing all visible design workflows."""
        return DesignWorkflowCollection(project_id=self.uid, session=self.session)

    @property
    def datasets(self) -> DatasetCollection:
        """Return a resource representing all visible datasets."""
        return DatasetCollection(self.uid, self.session)

    @property
    def tables(self) -> GemTableCollection:
        """Return a resource representing all visible Tables."""
        return GemTableCollection(self.uid, self.session)

    @property
    def property_templates(self) -> PropertyTemplateCollection:
        """Return a resource representing all property templates in this dataset."""
        return PropertyTemplateCollection(self.uid, None, self.session)

    @property
    def condition_templates(self) -> ConditionTemplateCollection:
        """Return a resource representing all condition templates in this dataset."""
        return ConditionTemplateCollection(self.uid, None, self.session)

    @property
    def parameter_templates(self) -> ParameterTemplateCollection:
        """Return a resource representing all parameter templates in this dataset."""
        return ParameterTemplateCollection(self.uid, None, self.session)

    @property
    def material_templates(self) -> MaterialTemplateCollection:
        """Return a resource representing all material templates in this dataset."""
        return MaterialTemplateCollection(self.uid, None, self.session)

    @property
    def measurement_templates(self) -> MeasurementTemplateCollection:
        """Return a resource representing all measurement templates in this dataset."""
        return MeasurementTemplateCollection(self.uid, None, self.session)

    @property
    def process_templates(self) -> ProcessTemplateCollection:
        """Return a resource representing all process templates in this dataset."""
        return ProcessTemplateCollection(self.uid, None, self.session)

    @property
    def process_runs(self) -> ProcessRunCollection:
        """Return a resource representing all process runs in this dataset."""
        return ProcessRunCollection(self.uid, None, self.session)

    @property
    def measurement_runs(self) -> MeasurementRunCollection:
        """Return a resource representing all measurement runs in this dataset."""
        return MeasurementRunCollection(self.uid, None, self.session)

    @property
    def material_runs(self) -> MaterialRunCollection:
        """Return a resource representing all material runs in this dataset."""
        return MaterialRunCollection(self.uid, None, self.session)

    @property
    def ingredient_runs(self) -> IngredientRunCollection:
        """Return a resource representing all ingredient runs in this dataset."""
        return IngredientRunCollection(self.uid, None, self.session)

    @property
    def process_specs(self) -> ProcessSpecCollection:
        """Return a resource representing all process specs in this dataset."""
        return ProcessSpecCollection(self.uid, None, self.session)

    @property
    def measurement_specs(self) -> MeasurementSpecCollection:
        """Return a resource representing all measurement specs in this dataset."""
        return MeasurementSpecCollection(self.uid, None, self.session)

    @property
    def material_specs(self) -> MaterialSpecCollection:
        """Return a resource representing all material specs in this dataset."""
        return MaterialSpecCollection(self.uid, None, self.session)

    @property
    def ingredient_specs(self) -> IngredientSpecCollection:
        """Return a resource representing all ingredient specs in this dataset."""
        return IngredientSpecCollection(self.uid, None, self.session)

    @property
    def gemd(self) -> GEMDResourceCollection:
        """Return a resource representing all GEMD objects/templates in this dataset."""
        return GEMDResourceCollection(self.uid, None, self.session)

    @property
    def table_configs(self) -> TableConfigCollection:
        """Return a resource representing all Table Configs in the project."""
        return TableConfigCollection(self.uid, self.session)

    @use_teams("project.make_public", True)
    def publish(self, *, resource: Resource):
        """
        Publish a resource from a project to its encompassing team.

        In order to use the Resource in a different project,
        you should use project.pull_in_resource() to pull that resource
        into the other project.

        Parameters
        ----------
        resource: Resource
            The resource owned by this project, which will be published

        Returns
        -------
        bool
            Returns ``True`` if resource successfully published

        """
        resource_access = resource.access_control_dict()
        resource_type = resource_access["type"]
        self.session.checked_post(
            f"{self._path()}/published-resources/{resource_type}/batch-publish",
            version='v3', json={'ids': [resource_access["id"]]})
        return True

    @use_teams("project.make_private", True)
    def un_publish(self, *, resource: Resource):
        """
        Un-publish a resource from a project from its encompassing team.

        Parameters
        ----------
        resource: Resource
            The resource owned by this project, which will be un-published

        Returns
        -------
        bool
            Returns ``True`` if resource successfully un-published

        """
        resource_access = resource.access_control_dict()
        resource_type = resource_access["type"]
        self.session.checked_post(
            f"{self._path()}/published-resources/{resource_type}/batch-un-publish",
            version='v3', json={'ids': [resource_access["id"]]})
        return True

    @use_teams("project.make_public", True)
    def pull_in_resource(self, *, resource: Resource):
        """
        Pull in a public resource from this project's team.

        Parameters
        ----------
        resource: Resource
            The resource owned by the encompassing team, which will be pulled in

        Returns
        -------
        bool
            Returns ``True`` if resource successfully pulled in

        """
        resource_access = resource.access_control_dict()
        resource_type = resource_access["type"]
        base_url = f'/teams/{self.team_id}{self._path()}'
        self.session.checked_post(
            f'{base_url}/outside-resources/{resource_type}/batch-pull-in',
            version='v3', json={'ids': [resource_access["id"]]})
        return True

    @use_teams("team.share")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.share")
    def share(self, *,
              resource: Optional[Resource] = None,
              project_id: Optional[Union[str, UUID]] = None,
              resource_type: Optional[str] = None,
              resource_id: Optional[str] = None
              ) -> Dict[str, str]:
        """Share a resource with another project.

        Parameters
        ----------
        resource: Resource
            The resource owned by this project, which will be shared
        project_id: Union[str, UUID]
            The id of the project with which to share the resource
        resource_type: Optional[str]
            [DEPRECATED] Please use ``resource`` instead
            The type of the resource to share. Must be one of DATASET, MODULE, USER,
            TABLE, OR TABLE_DEFINITION
        resource_id: Optional[str]
            [DEPRECATED] Please use ``resource`` instead
            The id of the resource to share

        """
        resource_dict = None
        if resource is not None:
            resource_dict = resource.access_control_dict()
        if resource_type is not None and resource_id is not None:
            warn("Asset sharing through resource_type and resource_id is deprecated. Please pass "
                 "the resource to share instead.", DeprecationWarning)
            if resource_dict is not None:
                raise ValueError("Cannot specify resource to share and also specify the "
                                 "resource type and id")
            resource_dict = {"type": resource_type, "id": resource_id}
        if resource_dict is None:
            raise ValueError("Must specify resource to share or specify the resource type and id")
        return self.session.post_resource(f"{self._path()}/share", {
            "project_id": str(project_id),
            "resource": resource_dict
        })

    @use_teams("project.publish")
    @v1_deprecation_warn()
    def transfer_resource(self, *, resource: Resource,
                          receiving_project_uid: Union[str, UUID]) -> bool:
        """
        Transfer ownership of a resource.

        The new owner of the the supplied resource becomes the project
        with ``uid == receiving_project_uid``.

        Parameters
        ----------
        resource: Resource
            The resource owned by this project, which will get transferred to
            the project with ``uid == receiving_project_uid``.
        receiving_project_uid: Union[string, UUID]
            The uid of the project to which the resource will be transferred.

        Returns
        -------
        bool
            Returns ``True`` upon successful resource transfer.

        """
        try:
            self.session.checked_post(f"{self._path()}/transfer-resource", {
                "to_project_id": str(receiving_project_uid),
                "resource": resource.access_control_dict()})
        except AttributeError:  # If _resource_type is not implemented
            raise RuntimeError(f"Resource of type  {resource.__class__.__name__} "
                               f"cannot be made transferred")

        return True

    @use_teams("project.publish")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use project.publish")
    def make_public(self, resource: Resource) -> bool:
        """
        Grant public access to a resource owned by this project.

        Parameters
        ----------
        resource: Resource
            An instance of a resource owned by this project (e.g., a dataset).

        Returns
        -------
        bool
            ``True`` if the action was performed successfully

        """
        try:
            self.session.checked_post(f"{self._path()}/make-public", {
                "resource": resource.access_control_dict()
            })
        except AttributeError:  # If _resource_type is not implemented
            raise RuntimeError(f"Resource of type  {resource.__class__.__name__} "
                               f"cannot be made public")
        return True

    @use_teams("project.un_publish")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use project.un_publish")
    def make_private(self, resource: Resource) -> bool:
        """
        Remove public access for a resource owned by this project.

        Parameters
        ----------
        resource: Resource
            An instance of a resource owned by this project (e.g., a dataset).

        Returns
        -------
        bool
            ``True`` if the action was performed successfully

        """
        try:
            self.session.checked_post(f"{self._path()}/make-private", {
                "resource": resource.access_control_dict()
            })
        except AttributeError:  # If _resource_type is not implemented
            raise RuntimeError(f"Resource of type  {resource.__class__.__name__} "
                               f"cannot be made private")
        return True

    def creator(self) -> str:
        """
        Return the creator of this project.

        Returns
        -------
        str
            The email of the creator of this resource.

        """
        if self.session._accounts_service_v3:
            raise NotImplementedError("Not available")
        email = self.session.get_resource(f"{self._path()}/creator")["email"]
        return email

    def owned_dataset_ids(self) -> List[str]:
        """
        List all the ids of the datasets owned by the current project.

        Returns
        -------
        List[str]
            The ids of the modules owned by current project

        """
        if self.session._accounts_service_v3:
            query_params = {"userId": "", "domain": self._path(), "action": "WRITE"}
            dataset_ids = self.session.get_resource("/DATASET/authorized-ids",
                                                    params=query_params,
                                                    version="v3")['ids']
        else:
            dataset_ids = self.session.get_resource(f"{self._path()}/dataset_ids")["dataset_ids"]
        return dataset_ids

    def owned_table_ids(self) -> List[str]:
        """
        List all the ids of the tables owned by the current project.

        Returns
        -------
        List[str]
            The ids of the tables owned by current project

        """
        if self.session._accounts_service_v3:
            raise NotImplementedError("Not available")
        table_ids = self.session.get_resource(f"{self._path()}/table_ids")["table_ids"]
        return table_ids

    def owned_table_config_ids(self) -> List[str]:
        """
        List all the ids of the table configs owned by the current project.

        Returns
        -------
        List[str]
            The ids of the table configs owned by current project

        """
        if self.session._accounts_service_v3:
            raise NotImplementedError("Not available")
        result = self.session.get_resource(f"{self._path()}/table_definition_ids")
        return result["table_definition_ids"]

    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.list_members", True)
    def list_members(self) -> Union[List[ProjectMember], List["TeamMember"]]:
        """
        List all of the members in the current project.

        Returns
        -------
        List[ProjectMember] | List[TeamMember]
            The members of the current project, or the members of the team
            containing the project if teams have been released.

        """
        from citrine.resources.team import TeamCollection

        if self.session._accounts_service_v3:
            team_collection = TeamCollection(self.session)
            parent_team = team_collection.get(self.team_id)
            return parent_team.list_members()
        else:
            members = self.session.get_resource(f"{self._path()}/users")["users"]
            return [ProjectMember(user=User.build(m),
                                  project=self,
                                  role=m["role"])
                    for m in members]

    @use_teams("team.update_user_action")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.update_user_action")
    def update_user_role(self, *, user_uid: Union[str, UUID], role: ROLES, actions: ACTIONS = []):
        """
        Update a User's role and action permissions in the Project.

        Valid roles are ``MEMBER`` or ``LEAD``.

        ``WRITE`` is the only action available for specification.

        Returns
        -------
        bool
            Returns ``True`` if user role successfully updated

        """
        self.session.checked_post(self._path() + format_escaped_url("/users/{}", user_uid),
                                  {'role': role, 'actions': actions})
        return True

    @use_teams("team.add_user")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.add_user")
    def add_user(self, user_uid: Union[str, UUID]):
        """
        Add a User to a Project.

        Adds User with ``MEMBER`` role to the Project.
        Use the ``update_user_rule`` method to change a User's role.

        Returns
        -------
        bool
            Returns ``True`` if user successfully added

        """
        self.session.checked_post(self._path() + format_escaped_url("/users/{}", user_uid),
                                  {'role': MEMBER, 'actions': []})
        return True

    @use_teams("team.remove_user")
    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.remove_user")
    def remove_user(self, user_uid: Union[str, UUID]) -> bool:
        """
        Remove a User from a Project.

        Returns
        -------
        bool
            Returns ``True`` if user successfully removed

        """
        self.session.checked_delete(
            self._path() + format_escaped_url("/users/{}", user_uid)
        )
        return True

    def gemd_batch_delete(
            self,
            id_list: List[Union[LinkByUID, UUID, str, BaseEntity]],
            *,
            timeout: float = 2 * 60,
            polling_delay: float = 1.0
    ) -> List[Tuple[LinkByUID, ApiError]]:
        """
        Remove a set of GEMD objects.

        You may provide GEMD objects that reference each other, and the objects
        will be removed in the appropriate order.

        A failure will be returned if the object cannot be deleted due to an external
        reference.

        You must have Write access on the associated datasets for each object.

        Parameters
        ----------
        id_list: List[Union[LinkByUID, UUID, str, BaseEntity]]
            A list of the IDs of data objects to be removed. They can be passed
            as a LinkByUID tuple, a UUID, a string, or the object itself. A UUID
            or string is assumed to be a Citrine ID, whereas a LinkByUID or
            BaseEntity can also be used to provide an external ID.
        timeout: float
            Amount of time to wait on the job (in seconds) before giving up. Defaults
            to 2 minutes. Note that this number has no effect on the underlying job
            itself, which can also time out server-side.
        polling_delay: float
            How long to delay between each polling retry attempt (in seconds).

        Returns
        -------
        List[Tuple[LinkByUID, ApiError]]
            A list of (LinkByUID, api_error) for each failure to delete an object.
            Note that this method doesn't raise an exception if an object fails to be
            deleted.

        """
        return _async_gemd_batch_delete(id_list, self.uid, self.session, None,
                                        timeout=timeout, polling_delay=polling_delay)


class ProjectCollection(Collection[Project]):
    """
    Represents the collection of all projects as well as the resources belonging to it.

    Parameters
    ----------
    session: Session, optional
        The Citrine session used to connect to the database.

    """

    _path_template = '/projects'
    _individual_key = 'project'
    _collection_key = 'projects'
    _resource = Project

    @property
    def _api_version(self):
        return 'v3' if self.session._accounts_service_v3 else 'v1'

    def __init__(self, session: Session, *, team_id: Optional[UUID] = None):
        self.session = session
        self.team_id = team_id

    def build(self, data) -> Project:
        """
        Build an individual project from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the project.

        Return
        -------
        Project
            The project created from data.

        """
        project = Project.build(data)
        project.session = self.session
        if self.team_id is not None:
            project.team_id = self.team_id
        return project

    def _register_in_team(self, name: str, *, description: Optional[str] = None):
        if self.team_id is None:
            raise NotImplementedError("Please use team.projects")
        path = format_escaped_url('teams/{team_id}/projects', team_id=self.team_id)
        project = Project(name, description=description)
        try:
            data = self.session.post_resource(path, project.dump(), version=self._api_version)
            data = data[self._individual_key]
            return self.build(data)
        except NonRetryableException as e:
            raise ModuleRegistrationFailedException(project.__class__.__name__, e)

    @v1_deprecation_warn("This method will be deprecated with the Team's release. \
        Once Teams are released, use team.projects.register", True)
    def register(self, name: str, *, description: Optional[str] = None) -> Project:
        """
        Create and upload new project.

        Parameters
        ----------
        name: str
            Name of the project to be created.
        description: str
            Long-form description of the project to be created.

        """
        if self.session._accounts_service_v3:
            return self._register_in_team(name, description=description)
        else:
            return super().register(Project(name, description=description))

    def list(self, *,
             page: Optional[int] = None,
             per_page: int = 1000) -> Iterator[Project]:
        """
        List projects using pagination.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 1000.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.

        Returns
        -------
        Iterator[Project]
            Projects in this collection.

        """
        if self.session._accounts_service_v3:
            return self._list_v3(page=page, per_page=per_page)
        else:
            return super().list(page=page, per_page=per_page)

    def _list_v3(self, *, page: Optional[int] = None, per_page: int = 1000) -> Iterator[Project]:
        if self.team_id is None:
            path = '/projects'
        else:
            path = format_escaped_url('/teams/{team_id}/projects', team_id=self.team_id)

        fetcher = partial(self._fetch_page, path=path)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def search_all(self, search_params: Optional[Dict]) -> Iterable[Dict]:
        """
        Search across all projects in a domain.

        There is no pagination on search_all.

         This is compatible with accounts v3 only.

        Parameters
        ----------
        search_params: dict, optional
            A ``dict`` representing the body of the post request that will be sent to the search
            endpoint to filter the results, e.g.,

            .. code:: python

                {
                    "name": {
                        "value": "Polymer Project",
                        "search_method": "EXACT"
                    },
                    "description": {
                        "value": "polymer chain length",
                        "search_method": "SUBSTRING"
                    },
                }

            The ``dict`` can contain any combination of (one or all) search specifications for the
            name, description, and status fields of a project. For each parameter specified, the
            ``"value"`` to match, as well as the ``"search_method"`` must be provided.
            The available ``search_methods`` are ``"SUBSTRING"`` and ``"EXACT"``. The example above
            demonstrates the input necessary to list projects with the exact name
            ``"Polymer Project"`` and descriptions including the phrase ``"polymer chain length"``.


        Returns
        -------
        Iterable[Dict]
            Projects in this collection.

        """
        collections = []
        path = self._get_path() + '/search'
        query_params = {'userId': ""}

        json = {} if search_params is None else {'search_params': search_params}

        data = self.session.post_resource(path,
                                          params=query_params,
                                          json=json,
                                          version=self._api_version)

        if self._collection_key is not None:
            collections = data[self._collection_key]

        return collections

    def search(self, *, search_params: Optional[dict] = None,
               per_page: int = 1000) -> Iterable[Project]:
        """
        Search for projects matching the desired name or description.

        You can search for projects matching the desired name or description by either exact match
        or substring match, as specified by the search_params argument. Defaults to no search
        criteria.

        Like ``list``, this method allows for pagination. This differs from the list function,
        because it makes a POST request to resourceType/search with search fields in a post body.

        Leaving page and per_page as default values will yield all elements in the collection,
        paginating over all available pages.

        Leaving ``search_params`` as its default value will return mimic the behavior of
        a full list with no search parameters.

        Parameters
        ----------
        search_params: dict, optional
            A ``dict`` representing the body of the post request that will be sent to the search
            endpoint to filter the results, e.g.,

            .. code:: python

                {
                    "name": {
                        "value": "Polymer Project",
                        "search_method": "EXACT"
                    },
                    "description": {
                        "value": "polymer chain length",
                        "search_method": "SUBSTRING"
                    },
                }

            The ``dict`` can contain any combination of (one or all) search specifications for the
            name, description, and status fields of a project. For each parameter specified, the
            ``"value"`` to match, as well as the ``"search_method"`` must be provided.
            The available ``search_methods`` are ``"SUBSTRING"`` and ``"EXACT"``. The example above
            demonstrates the input necessary to list projects with the exact name
            ``"Polymer Project"`` and descriptions including the phrase ``"polymer chain length"``.

        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter is used when
            making requests to the backend service.  If the page parameter is specified it limits
            the maximum number of elements in the response.

        Returns
        -------
        Iterable[Project]
            Projects in this collection.

        """
        if self.session._accounts_service_v3:
            return self._build_collection_elements(self.search_all(search_params))
        # To avoid setting default to {} -> reduce mutation risk, and to make more extensible

        return self._paginator.paginate(page_fetcher=self._fetch_page_search,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page,
                                        search_params=search_params)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """
        [ALPHA] Delete a particular project.

        Only empty projects can be deleted.
        If the project is not empty, then the Response will contain a list of all of the project's
        resources. These must be deleted before the project can be deleted.
        """
        return super().delete(uid)

    def update(self, model: Project) -> Project:
        """Projects cannot be updated."""
        raise NotImplementedError("Project update is not supported at this time.")

    def _fetch_page_search(self, page: Optional[int] = None,
                           per_page: Optional[int] = None,
                           search_params: Optional[dict] = None) -> Tuple[Iterable[dict], str]:
        """
        Fetch resources that match the supplied search parameters.

        Fetches resources that match the supplied ``search_params``, by calling ``_fetch_page``
        with ``checked_post``, the path to the POST resource-type/search endpoint, any pagination
        parameters, and the request body to the search endpoint.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is the first page, which is 1.
        per_page: int, optional
            Max number of results to return. Default is 20.
        search_params: dict, optional
            A ``dict`` representing a request body that could be sent to a POST request. The "json"
            field should be passed as the key for the outermost ``dict``, with its value the
            request body, so that we can easily unpack the keyword argument when it gets passed to
            ``fetch_func``, e.g., ``{'name': {'value': 'Project', 'search_method': 'SUBSTRING'} }``

        Returns
        -------
        Iterable[dict]
            Elements in this collection.
        str
            The next uri if one is available, empty string otherwise

        """
        # Making 'json' the key of the outermost dict, so that search_params can be passed
        # directly to the function making the request with keyword expansion
        json_body = {} if search_params is None else {'json': {'search_params': search_params}}

        path = self._get_path() + "/search"

        return self._fetch_page(path=path, fetch_func=self.session.post_resource,
                                page=page, per_page=per_page,
                                json_body=json_body)



import warnings
from copy import deepcopy
from typing import Callable, Union, Iterable, Optional, Tuple
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._session import Session
from citrine._utils.functions import migrate_deprecated_argument, format_escaped_url
from citrine.informatics.workflows import DesignWorkflow
from citrine.resources.response import Response
from functools import partial


class DesignWorkflowCollection(Collection[DesignWorkflow]):
    """A collection of DesignWorkflows."""

    _path_template = '/projects/{project_id}/design-workflows'
    _individual_key = None
    _collection_key = 'response'
    _resource = DesignWorkflow
    _api_version = "v2"

    def __init__(self, project_id: UUID, session: Session, branch_id: UUID = None):
        self.project_id: UUID = project_id
        self.session: Session = session
        self.branch_id: UUID = branch_id

    def register(self, model: DesignWorkflow) -> DesignWorkflow:
        """
        Upload a new design workflow.

        The model's branch ID is ignored in favor of this collection's. If this
        collection has a null branch ID, then the model is uploaded to the v1
        endpoint and retrieved separately, to ensure all details are included.

        Parameters
        ----------
        model: DesignWorkflow
            The design workflow to be uploaded.

        Return
        -------
        DesignWorkflow
            The newly created design workflow.

        """
        if self.branch_id is None:
            # Prior to the introduction of branches, partial design workflows were not supported
            # at all. As such, their use without a branch continues to be an error.
            if model.predictor_id is None or model.design_space_id is None:
                raise ValueError("A design workflow without a predictor ID and/or a design space "
                                 "ID must be registered to a specific branch.")

            # There are a number of contexts in which hitting design workflow endpoints without a
            # branch ID is valid, so only this particular usage is deprecated.
            msg = ('Creating a design workflow without a branch is deprecated as of 1.19.0 and '
                   'will be removed in 2.0.0. Branches are a concept introduced in the CP2 '
                   'version of the Citrine Platform. To learn more, see our documentation at '
                   'https://citrineinformatics.github.io/citrine-python/workflows/design_workflows.html#branches')  # noqa
            warnings.warn(msg, category=DeprecationWarning)

            # To create a design workflow without providing a branch ID, we need to hit the v1
            # API, then do a GET to grab the ID of the branch that was created automatically.
            v1 = _DesignWorkflowCollectionV1(self.project_id, self.session)
            # Passing in the subclass instance as self avoids infinite recursion.
            created_dw = super(DesignWorkflowCollection, v1).register(model)
            return super().get(created_dw.uid)
        else:
            # branch_id is in the body of design workflow endpoints, so it must be serialized.
            # This means the collection branch_id might not match the workflow branch_id. The
            # collection should win out, since the user is explicitly referencing the branch
            # represented by this collection.
            # To avoid modifying the parameter, and to ensure the only change is the branch_id, we
            # deepcopy, modify, then register it.
            model_copy = deepcopy(model)
            model_copy.branch_id = self.branch_id
            return super().register(model_copy)

    def build(self, data: dict) -> DesignWorkflow:
        """
        Build an individual design workflow from a dictionary.

        Parameters
        ----------
        data: dict
            A dictionary representing the design workflow.

        Return
        -------
        DesignWorkflow
            The design workflow created from data.

        """
        workflow = DesignWorkflow.build(data)
        workflow._session = self.session
        workflow.project_id = self.project_id
        return workflow

    def update(self, model: DesignWorkflow) -> DesignWorkflow:
        """Update a design workflow.

        Identifies the workflow by the model's uid. It must have a branch_id, and if this
        collection also has a branch_id, they must match. Prefer updating a workflow through
        Project.design_workflows.update.

        Parameters
        ----------
        model: DesignWorkflow
            The design workflow values that are desired, identified by the uid.

        Return
        -------
        DesignWorkflow
            The design workflow resulting from the update.

        """
        if self.branch_id is not None:
            if self.branch_id != model.branch_id:
                raise ValueError('To move a design workflow to another branch, please use '
                                 'Project.design_workflows.update')

        if model.branch_id is None:
            raise ValueError('Cannot update a design workflow unless its branch_id is set.')

        return super().update(model)

    def archive(self, uid: Union[UUID, str] = None, workflow_id: Union[UUID, str] = None):
        """Archive a design workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to archive
        workflow_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", workflow_id, "workflow_id")
        url = format_escaped_url(self._path_template, project_id=self.project_id) \
            + format_escaped_url("/{}/archive", uid)
        self.session.put_resource(url, {}, version=self._api_version)

    def restore(self, uid: Union[UUID, str] = None, workflow_id: [UUID, str] = None):
        """Restore an archived design workflow.

        Parameters
        ----------
        uid: Union[UUID, str]
            Unique identifier of the workflow to restore
        workflow_id: Union[UUID, str]
            [DEPRECATED] please use uid instead

        """
        uid = migrate_deprecated_argument(uid, "uid", workflow_id, "workflow_id")
        url = format_escaped_url(self._path_template, project_id=self.project_id) \
            + format_escaped_url("/{}/restore", uid)
        self.session.put_resource(url, {}, version=self._api_version)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Design Workflows cannot be deleted; they can be archived instead."""
        raise NotImplementedError(
            "Design Workflows cannot be deleted; they can be archived instead.")

    def list_archived(self,
                      *,
                      page: Optional[int] = None,
                      per_page: int = 500) -> Iterable[DesignWorkflow]:
        """List archived Design Workflows."""
        fetcher = partial(self._fetch_page, additional_params={"filter": "archived eq 'true'"})
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def _fetch_page(self,
                    path: Optional[str] = None,
                    fetch_func: Optional[Callable[..., dict]] = None,
                    page: Optional[int] = None,
                    per_page: Optional[int] = None,
                    json_body: Optional[dict] = None,
                    additional_params: Optional[dict] = None,
                    ) -> Tuple[Iterable[dict], str]:
        params = additional_params or {}
        params["branch"] = self.branch_id
        return super()._fetch_page(path=path,
                                   fetch_func=fetch_func,
                                   page=page,
                                   per_page=per_page,
                                   json_body=json_body,
                                   additional_params=params)


class _DesignWorkflowCollectionV1(DesignWorkflowCollection):
    """A small proxy class to direct all calls to the v1 endpoints."""

    _api_version = "v1"



"""Tools for working with Constraints."""

from citrine._serialization.polymorphic_serializable import PolymorphicSerializable

__all__ = ['Constraint']


class Constraint(PolymorphicSerializable['Constraint']):
    """A Citrine Constraint places restrictions on a design space.

    Abstract type that returns the proper type given a serialized dict.

    """

    _response_key = None

    @classmethod
    def get_type(cls, data):
        """Return the subtype."""
        from .ingredient_count_constraint import IngredientCountConstraint
        from .ingredient_fraction_constraint import IngredientFractionConstraint
        from .label_fraction_constraint import LabelFractionConstraint
        from .scalar_range_constraint import ScalarRangeConstraint
        from .categorical_constraint import AcceptableCategoriesConstraint
        return {
            'Categorical': AcceptableCategoriesConstraint,  # Kept for backwards compatibility.
            'AcceptableCategoriesConstraint': AcceptableCategoriesConstraint,
            'IngredientCountConstraint': IngredientCountConstraint,
            'IngredientFractionConstraint': IngredientFractionConstraint,
            'LabelFractionConstraint': LabelFractionConstraint,
            'ScalarRange': ScalarRangeConstraint,
        }[data['type']]



import warnings
from typing import List, Optional

from citrine._rest.engine_resource import EngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictors import Predictor

__all__ = ['AutoMLPredictor']


class AutoMLPredictor(EngineResource['AutoMLPredictor'], Predictor):
    """A predictor interface that builds a single ML model.

    The model uses the set of inputs to predict the output(s).
    Only one machine learning model is built.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    inputs: list[Descriptor]
        Descriptors that represent inputs to the model
    output: Descriptor
        [DEPRECATED] Please use ``outputs`` instead
        A single Descriptor that represents the output of the model
    outputs: list[Descriptor]
        Descriptors that represents the output(s) of the model.
        Currently, only one output Descriptor is supported.
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. De-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    inputs = _properties.List(_properties.Object(Descriptor), 'data.instance.inputs')
    outputs = _properties.List(_properties.Object(Descriptor), 'data.instance.outputs')
    training_data = _properties.List(_properties.Object(DataSource),
                                     'data.instance.training_data', default=[])

    typ = _properties.String('data.instance.type', default='AutoML', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 output: Descriptor = None,
                 outputs: List[Descriptor] = None,
                 inputs: List[Descriptor],
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.inputs: List[Descriptor] = inputs
        self.training_data: List[DataSource] = training_data or []

        if output is not None:
            msg = ('The "output" parameter is deprecated as of 1.24.0 and will be removed in '
                   '2.0.0. Please use the "outputs" field instead.')
            warnings.warn(msg, category=DeprecationWarning)
            if outputs is not None:
                raise ValueError('Found values for "outputs" and the deprecated "output". Please '
                                 'provide only "outputs".')

        self.outputs: List[Descriptor] = []
        if outputs is not None:
            self.outputs = outputs
        elif output is not None:
            self.outputs = [output]

    @property
    def output(self) -> Optional[Descriptor]:
        """[DEPRECATED] Get the first output descriptor."""
        msg = ('The "output" field is deprecated as of 1.24.0 and will be removed in 2.0.0. '
               'Please use the "outputs" field instead.')
        warnings.warn(msg, category=DeprecationWarning)
        return self.outputs[0] if self.outputs else None

    @output.setter
    def output(self, value):
        """[DEPRECATED] Set the output descriptor."""
        msg = ('The "output" field is deprecated as of 1.24.0 and will be removed in 2.0.0. '
               'Please use the "outputs" field instead.')
        warnings.warn(msg, category=DeprecationWarning)
        if not isinstance(value, Descriptor):
            # Mirroring the error emited by serialization.
            raise ValueError(f"{value} is not one of valid types: <class 'Descriptor'> for output")

        self.outputs = [value]

    def __str__(self):
        return '<AutoMLPredictor {!r}>'.format(self.name)



from logging import getLogger
from typing import Optional, Iterable, Dict, Tuple, Callable, Union
from uuid import UUID


logger = getLogger(__name__)


class Pageable():
    """Class that allows paging."""

    _collection_key: str = NotImplemented
    _api_version: str = "v1"

    def _get_path(self, uid: Optional[Union[UUID, str]] = None,
                  ignore_dataset: Optional[bool] = False) -> str:
        """Construct a url from __base_path__ and, optionally, id."""
        raise NotImplementedError  # pragma: no cover

    def _fetch_page(self,
                    path: Optional[str] = None,
                    fetch_func: Optional[Callable[..., dict]] = None,
                    page: Optional[int] = None,
                    per_page: Optional[int] = None,
                    json_body: Optional[dict] = None,
                    additional_params: Optional[dict] = None,
                    ) -> Tuple[Iterable[dict], str]:
        """
        Fetch visible elements.  This does not handle pagination.

        Method can be used with any function that fetches a list of resources.

        This method will return the first page of results using the default page/per_page behavior
        of the backend service.  Specify page/per_page to override these defaults which are passed
        to the backend service.

        Parameters
        ---------
        path: str, optional
            The path for the endpoint that will be called to fetch the resources. Will default to
            root path
        fetch_func: Callable[..., dict], optional
            The function that will make the official request that returns the list of resources ie.
            (checked_post, etc.). Will default to get_resource
        page: int, optional
            The "page" of results to list. Default is the first page, which is 1.
        per_page: int, optional
            Max number of results to return. Default is 20.
        json_body: dict, optional
            A dict representing a request body that could be sent to a POST request. The "json"
            field should be passed as the key for the outermost dict, with its value the request
            body, so that we can easily unpack the keyword argument when it gets passed to
            fetch_func.
            ie.
            {'json':
                {'search_params': {'name': {'value': 'Project', 'search_method': 'SUBSTRING'}}}
            }
        additional_params: dict, optional
            A dict that allows extra parameters to be added to the request parameters

        Returns
        -------
        Iterable[dict]
            Elements in this collection.
        str
            The next uri if one is available, empty string otherwise

        """
        # To avoid setting defaults -> reduce mutation risk, and to make more extensible
        path = self._get_path() if path is None else path
        fetch_func = self.session.get_resource if fetch_func is None else fetch_func
        json_body = {} if json_body is None else json_body

        module_type = getattr(self, '_module_type', None)
        params = self._page_params(page, per_page, module_type)
        params.update(additional_params or {})

        data = fetch_func(path, params=params, version=self._api_version, **json_body)

        try:
            next_uri = data.get('next', "")
        except AttributeError:
            next_uri = ""

        # A 'None' collection key implies response has a top-level array
        # of 'ResourceType'
        # TODO: Unify backend return values
        if self._collection_key is None:
            collection = data
        else:
            collection = data[self._collection_key]

        return collection, next_uri

    def _page_params(self,
                     page: Optional[int],
                     per_page: Optional[int],
                     module_type: Optional[str] = None) -> Dict[str, int]:
        params = {}
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page
        if module_type is not None:
            params["module_type"] = module_type
        return params



from typing import TypeVar, Optional, Union
from uuid import UUID

from citrine._serialization.serializable import Serializable
from citrine._serialization import properties
from gemd.enumeration.base_enumeration import BaseEnumeration


class ResourceTypeEnum(BaseEnumeration):
    """The type of the resource; used for modifying access controls.

    * TEAM is a Team
    * PROJECT is a Project
    * DATASET is a Dataset
    * MODULE is a Module: a Predictor, Design Space, or Processor
    * USER is a user
    * TABLE is a GemTable
    * TABLE_DEFINITION is a TableConfig

    """

    TEAM = "TEAM"
    PROJECT = "PROJECT"
    DATASET = "DATASET"
    MODULE = "MODULE"
    USER = "USER"
    TABLE = "TABLE"
    TABLE_DEFINITION = "TABLE_DEFINITION"


Self = TypeVar('Self', bound='Resource')


class Resource(Serializable[Self]):
    """Abstract class for representing individual REST resource."""

    _path_template: str = NotImplemented
    _response_key: Optional[str] = None
    _resource_type: ResourceTypeEnum = NotImplemented

    def access_control_dict(self) -> dict:
        """Return an access control entity representation of this resource. Internal use only."""
        return {
            "type": self._resource_type.value,
            "id": str(self.uid)
        }


class ResourceRef(Serializable['ResourceRef']):
    """A reference to a resource by UID."""

    # json key 'module_uid' is a legacy of when this object was only used for modules
    uid = properties.UUID('module_uid')

    def __init__(self, uid: Union[UUID, str]):
        self.uid = uid



from typing import List

from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.predictor_evaluator import PredictorEvaluator
from citrine.informatics.workflows.workflow import Workflow
from citrine.resources.predictor_evaluation_execution import PredictorEvaluationExecutionCollection
from citrine._rest.ai_resource_metadata import AIResourceMetadata

__all__ = ['PredictorEvaluationWorkflow']


class PredictorEvaluationWorkflow(Resource['PredictorEvaluationWorkflow'],
                                  Workflow, AIResourceMetadata):
    """A workflow that evaluations a predictor.

    Parameters
    ----------
    name: str
        name of the predictor evaluation workflow
    description: str
        the description of the predictor evaluation workflow
    evaluators: List[PredictorEvaluator]
        the list of evaluators to apply to the predictor

    """

    evaluators = properties.List(properties.Object(PredictorEvaluator), "evaluators")

    status_description = properties.String('status_description', serializable=False)
    """:str: more detailed description of the workflow's status"""
    typ = properties.String('type', default='PredictorEvaluationWorkflow', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str = "",
                 evaluators: List[PredictorEvaluator]):
        self.name: str = name
        self.description: str = description
        self.evaluators: List[PredictorEvaluator] = evaluators

    def __str__(self):
        return '<PredictorEvaluationWorkflow {!r}>'.format(self.name)

    @property
    def executions(self) -> PredictorEvaluationExecutionCollection:
        """Return a resource representing all visible executions of this workflow."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize execution without project reference!')
        return PredictorEvaluationExecutionCollection(
            project_id=self.project_id, session=self._session, workflow_id=self.uid)



# flake8: noqa
# The docstring includes many long links that violate flake8, and it's easier to noqa
# the whole file than to pick out the offending lines.
from typing import List, Optional

from citrine._rest.engine_resource import EngineResource
from citrine._serialization import properties as _properties
from citrine._utils.functions import migrate_deprecated_argument
from citrine.informatics.descriptors import Descriptor, MolecularStructureDescriptor
from citrine.informatics.predictors import Predictor

__all__ = ['MolecularStructureFeaturizer']


class MolecularStructureFeaturizer(EngineResource['MolecularStructureFeaturizer'], Predictor):
    """
    A featurizer for molecular structures, powered by CDK.

    The MolecularStructureFeaturizer will compute a configurable set of features on molecular
    structure data, e.g., SMILES or InChI strings.  The features are computed using the
    `Chemistry Development Kit (CDK) <https://cdk.github.io/>`_.  The features are configured
    using the ``features`` and ``excludes`` arguments, which accept either feature names or predefined
    aliases.

    The default is the `standard` alias, corresponding to eight features that are
    a good balance of cost and performance:

        - `AcidGroupCount  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AcidicGroupCountDescriptor.html>`_
        - `AtomCount  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/AtomContainer.html>`_
        - `AtomicPolarizability  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/charges/Polarizability.html>`_
        - `BondCount <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/AtomContainer.html>`_
        - `HBondAcceptorCount  <http://cdk.github.io/cdk/latest/docs/api/org/openscience/cdk/qsar/descriptors/molecular/HBondAcceptorCountDescriptor.html>`_
        - `HBondDonorCount  <http://cdk.github.io/cdk/2.2/docs/api/org/openscience/cdk/qsar/descriptors/molecular/HBondDonorCountDescriptor.html>`_
        - `MassAutocorr  <http://cdk.github.io/cdk/2.2/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AutocorrelationDescriptorMass.html>`_
        - `PolarizabilityAutocorr  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AutocorrelationDescriptorPolarizability.html>`_

    The ``extended`` alias includes more features that may improve model performance but are slower and may dilute the signal in the features.
    It includes the ``standard`` set and:

        - `ALOGP  <http://cdk.github.io/cdk/1.4/docs/api/org/openscience/cdk/qsar/descriptors/molecular/ALOGPDescriptor.html>`_
        - `AromaticAtomCount  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AromaticAtomsCountDescriptor.html>`_
        - `AromaticBondCount  <http://cdk.github.io/cdk/2.2/docs/api/org/openscience/cdk/aromaticity/Aromaticity.html>`_
        - `BasicGroupCount  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/BasicGroupCountDescriptor.html>`_
        - `BPol  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/BPolDescriptor.html>`_
        - `CarbonTypes  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/CarbonTypesDescriptor.html>`_
        - `EccentricConnectivityIndex  <http://cdk.github.io/cdk/2.2/docs/api/org/openscience/cdk/qsar/descriptors/molecular/EccentricConnectivityIndexDescriptor.html>`_
        - `FMF  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/FMFDescriptor.html>`_
        - `FragmentComplexity  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/FragmentComplexityDescriptor.html>`_
        - `HybridizationRatioDescriptor  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/HybridizationRatioDescriptor.html>`_
        - `KappaShapeIndices  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/KappaShapeIndicesDescriptor.html>`_
        - `KierHallSmarts  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/KierHallSmartsDescriptor.html>`_
        - `LargestChain  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/LargestChainDescriptor.html>`_
        - `LargestPiSystem  <http://cdk.github.io/cdk/2.2/docs/api/org/openscience/cdk/qsar/descriptors/molecular/LargestPiSystemDescriptor.html>`_
        - `MannholdLogP  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/MannholdLogPDescriptor.html>`_
        - `MDE  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/MDEDescriptor.html>`_
        - `PetitjeanNumber  <http://cdk.github.io/cdk/latest/docs/api/org/openscience/cdk/qsar/descriptors/molecular/PetitjeanNumberDescriptor.html>`_
        - `RotatableBondsCount  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/RotatableBondsCountDescriptor.html>`_
        - `RuleOfFive  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/RuleOfFiveDescriptor.html>`_
        - `TopologicalSurfaceArea  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/FractionalPSADescriptor.html>`_
        - `VertexAdjacencyMagnitude  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/VAdjMaDescriptor.html>`_
        - `Weight  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/WeightDescriptor.html>`_
        - `WienerNumbers  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/WienerNumbersDescriptor.html>`_
        - `XLogP  <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/XLogPDescriptor.html>`_
        - `ZagrebIndex  <http://cdk.github.io/cdk/latest/docs/api/org/openscience/cdk/qsar/descriptors/molecular/ZagrebIndexDescriptor.html>`_

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    descriptor: MolecularStructureDescriptor
        the descriptor to featurize
    features: List[str]
        the list of features to compute, either by name or by group alias.
    excludes: List[str]
        list of features to exclude (accepts same set of values as features). The final set
        of outputs generated by the predictor is set(features) - set(excludes).

    """

    input_descriptor = _properties.Object(Descriptor, 'data.instance.descriptor')
    features = _properties.List(_properties.String, 'data.instance.features')
    excludes = _properties.List(_properties.String, 'data.instance.excludes')

    typ = _properties.String('data.instance.type', default='MoleculeFeaturizer', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 input_descriptor: MolecularStructureDescriptor = None,
                 features: Optional[List[str]] = None,
                 excludes: Optional[List[str]] = None,
                 descriptor: MolecularStructureDescriptor = None):
        self.name: str = name
        self.description: str = description
        input_descriptor = migrate_deprecated_argument(
            input_descriptor, "input_descriptor", descriptor, "descriptor"
        )
        self.input_descriptor = input_descriptor
        self.features = features if features is not None else ["standard"]
        self.excludes = excludes if excludes is not None else []

    def __str__(self):
        return '<MolecularStructureFeaturizer {!r}>'.format(self.name)



import inspect
from functools import wraps
import os
from pathlib import Path
from typing import Any, Optional, Union
from urllib.parse import urlparse, quote
from warnings import warn

from gemd.entity.link_by_uid import LinkByUID


def get_object_id(object_or_id):
    """Extract the citrine id from a data concepts object or LinkByUID."""
    from gemd.entity.attribute.base_attribute import BaseAttribute
    from citrine.resources.data_concepts import DataConcepts, CITRINE_SCOPE

    if isinstance(object_or_id, BaseAttribute):
        raise ValueError("Attributes do not have ids.")
    if isinstance(object_or_id, DataConcepts):
        citrine_id = object_or_id.uids.get(CITRINE_SCOPE)
        return citrine_id
    if isinstance(object_or_id, LinkByUID):
        if object_or_id.scope == CITRINE_SCOPE:
            return object_or_id.id
        raise ValueError("LinkByUID must be scoped to citrine scope {}, "
                         "instead is {}".format(CITRINE_SCOPE, object_or_id.scope))
    raise TypeError("{} must be a data concepts object or LinkByUID".format(object_or_id))


def validate_type(data_dict: dict, type_name: str) -> dict:
    """Ensure that dict has field 'type' with given value."""
    data_dict_copy = data_dict.copy()
    if 'type' in data_dict_copy:
        if data_dict_copy['type'] != type_name:
            raise Exception(
                "Object type must be {}, but was instead {}.".format(type_name, data_dict['type']))
    else:
        data_dict_copy['type'] = type_name

    return data_dict_copy


def scrub_none(json):
    """
    Recursively delete dictionary keys with the value ``None``.

    This action modifies the data structure in place.
    """
    if isinstance(json, dict):
        for key, value in list(json.items()):
            if value is None:
                del json[key]
            else:
                json[key] = scrub_none(json[key])
    elif isinstance(json, list):
        for idx, element in reversed(list(enumerate(json))):
            if element is not None:
                json[idx] = scrub_none(element)
    return json


def replace_objects_with_links(json: dict) -> dict:
    """For each top-level object in a dictionary, replace it with a Link, if appropriate."""
    for key, value in list(json.items()):
        json[key] = object_to_link(value)
    return json


def object_to_link(obj: Any) -> Any:
    """See if an object is a dictionary that can be converted into a Link, and if so, convert."""
    if isinstance(obj, dict):
        if 'type' in obj and 'uids' in obj and obj['type'] != LinkByUID.typ:
            return object_to_link_by_uid(obj)
        else:
            return replace_objects_with_links(obj)
    elif isinstance(obj, (tuple, list)):
        return [object_to_link(entry) for entry in obj]
    return obj


def object_to_link_by_uid(json: dict) -> dict:
    """Convert an object dictionary into a LinkByUID dictionary, if possible."""
    from citrine.resources.data_concepts import CITRINE_SCOPE
    if 'uids' in json:
        uids = json['uids']
        if not isinstance(uids, dict) or not uids:
            return json
        if CITRINE_SCOPE in uids:
            scope = CITRINE_SCOPE
        else:
            scope = next(iter(uids))
        this_id = uids[scope]
        return LinkByUID(scope, this_id).as_dict()
    else:
        return json


def rewrite_s3_links_locally(url: str, s3_endpoint_url: str = None) -> str:
    """
    Rewrite s3 links from localstack.

    When the tests are run inside Devkit, Localstack uses the hostname "localstack" and port
    4566 for the local S3 files URL. If you're hitting localhost from outside devkit, the
    host "localstack" won't resolve, but the port 4566 is still mapped to the host port (also
    on 4566), so we need to write the file links S3 endpoint hostname to "localhost".

    The caller should supply the correct S3 endpoint URL, eg: session.s3_endpoint_url
    """
    parsed_url = urlparse(url)

    if s3_endpoint_url is not None:
        # Given an explicit endpoint to use instead
        parsed_s3_endpoint = urlparse(s3_endpoint_url)
        return parsed_url._replace(scheme=parsed_s3_endpoint.scheme,
                                   netloc=parsed_s3_endpoint.netloc).geturl()
    else:
        # Else return the URL unmodified
        return url


def write_file_locally(content, local_path: Union[str, Path]):
    """Take content from remote and ensure path exists."""
    if isinstance(local_path, str):
        if len(os.path.split(local_path)[-1]) == 0:
            raise ValueError(f"A filename must be provided in the path ({local_path})")
        local_path = Path(local_path)

    # Resolve ~, .., and the like
    local_path = local_path.expanduser().resolve()
    if local_path.is_dir():
        raise ValueError(f"A filename must be provided in the path ({local_path})")

    local_path.parent.mkdir(parents=True, exist_ok=True)
    local_path.open(mode='wb').write(content)


def shadow_classes_in_module(source_module, target_module):
    """Shadow classes from a source to a target module, for backwards compatibility purposes."""
    for c in [cls for _, cls in inspect.getmembers(source_module, inspect.isclass) if
              cls.__module__ == source_module.__name__]:
        setattr(target_module, c.__qualname__, c)


def migrate_deprecated_argument(
        new_arg: Optional[Any],
        new_arg_name: str,
        old_arg: Optional[Any],
        old_arg_name: str
) -> Any:
    """
    Facilitates the migration of an argument's name.

    This method handles the situation in which a function has two arguments for the same thing,
    one old and one new. It ensures that only one of the two arguments is provided, throwing a
    ValueError is both/neither are provided. If the old version of the argument is provided,
    it throws a deprecation warning.

    Parameters
    ----------
    new_arg: Optional[Any]
        the value provided using the new argument (or None, if not provided)
    new_arg_name: str
        the new name of the argument (used for creating user-facing messages)
    old_arg: Optional[Any]
        the value provided using the old argument (or None, if not provided)
    old_arg_name: str
        the old name of the argument (used for creating user-facing messages)

    Returns
    -------
    Any
        the value of the argument to be used by the calling method

    """
    if old_arg is not None:
        warn(f"\'{old_arg_name}\' is deprecated in favor of \'{new_arg_name}\'",
             DeprecationWarning)
        if new_arg is None:
            return old_arg
        else:
            raise ValueError(f"Cannot specify both \'{new_arg_name}\' and \'{new_arg_name}\'")
    elif new_arg is None:
        raise ValueError(f"Please specify \'{new_arg_name}\'")
    return new_arg


def format_escaped_url(
        template: str,
        *args,
        **kwargs
) -> str:
    """
    Escape arguments with percent encoding and bind them to a template of a URL.

    This method takes a template string and binds the positional and keyword arguments
    passing it into a format statement.

    Parameters
    ----------
    template: str
        the `format` template to which the escaped arguments will be bound
    *args : Iterable[str]
        Other arguments
    **kwargs: Dict[str]
        Keyword arguments

    Returns
    -------
    str
        the formatted URL

    """
    return template.format(*[quote(str(x), safe='') for x in args],
                           **{k: quote(str(v), safe='') for (k, v) in kwargs.items()}
                           )


def use_teams(alt, negate=False):
    """
    Raises error with a redirect message if method is unavailable with(out) teams support.

    Parameters
    ----------
    alt: str
        the alternative method the user should use. Added to the Error message
    negate: bool
        Set to True if the error should be raised if teams are in use. Keep False
        if error should be raised if teams are not available.

    """
    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if negate:
                if not self.session._accounts_service_v3:
                    raise NotImplementedError(f"Not available, you may be looking for {alt}")
            else:
                if self.session._accounts_service_v3:
                    raise NotImplementedError(f"Not available, you may be looking for {alt}")
            return f(self, *args, **kwargs)

        return wrapper

    return decorator


def v1_deprecation_warn(
        msg="This method will be deprecated once teams are released", warn_with_teams=False):
    """
    Raises deprecation warning for methods that will be unavailable with teams release.

    Parameters
    ----------
    msg: str
        The deprecation message that will be printed in the warning.
    warn_with_teams: bool
        Whether or not to keep the deprecation warning once teams are released.

    """
    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if not self.session._accounts_service_v3 or warn_with_teams:
                warn(msg, DeprecationWarning)
            return f(self, *args, **kwargs)
        return wrapper

    return decorator



"""Top-level class for all object spec objects and collections thereof."""
from abc import ABC
from typing import TypeVar

from citrine.resources.data_objects import DataObject, DataObjectCollection


class ObjectSpec(DataObject, ABC):
    """
    An abstract object spec object.

    ObjectSpec must be extended along with `Resource`
    """


ObjectSpecResourceType = TypeVar("ObjectSpecResourceType", bound="ObjectSpec")


class ObjectSpecCollection(DataObjectCollection[ObjectSpecResourceType], ABC):
    """A collection of one kind of object spec object."""



"""A resource that represents a single module report."""
from citrine._rest.resource import Resource
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.reports import Report

from uuid import UUID


class ReportResource(Resource['ReportResource']):
    """Defines a resource for fetching reports from a module.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _path_template = '/projects/{project_id}/modules/{module_id}/report'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session = session

    def get(self, module_id: UUID) -> Report:
        """Gets a single report keyed on the module_id."""
        url_path = format_escaped_url(self._path_template,
                                      project_id=self.project_id,
                                      module_id=module_id)
        data = self.session.get_resource(url_path)
        report = Report.build(data)
        report.session = self.session
        return report



from typing import Set, Mapping

from citrine._rest.engine_resource import EngineResource
from citrine._serialization import properties as _properties
from citrine.informatics.descriptors import FormulationDescriptor, RealDescriptor
from citrine.informatics.predictors import Predictor

__all__ = ['IngredientsToFormulationPredictor']


class IngredientsToFormulationPredictor(
        EngineResource['IngredientsToFormulationPredictor'], Predictor):
    """[ALPHA] A predictor interface that constructs a formulation from ingredient quantities.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        description of the predictor
    output: FormulationDescriptor
        descriptor that represents the output formulation
    id_to_quantity: Mapping[str, RealDescriptor]
        Map from ingredient identifier to the descriptor that represents its quantity,
        e.g., ``{'water': RealDescriptor('water quantity', 0, 1, "")}``
    labels: Mapping[str, Set[str]]
        Map from each label to all ingredients assigned that label, when present in a mixture,
        e.g., ``{'solvent': {'water'}}``

    """

    output = _properties.Object(FormulationDescriptor, 'data.instance.output')
    id_to_quantity = _properties.Mapping(_properties.String, _properties.Object(RealDescriptor),
                                         'data.instance.id_to_quantity')
    labels = _properties.Mapping(_properties.String, _properties.Set(_properties.String),
                                 'data.instance.labels')

    typ = _properties.String('data.instance.type', default='IngredientsToSimpleMixture',
                             deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 output: FormulationDescriptor,
                 id_to_quantity: Mapping[str, RealDescriptor],
                 labels: Mapping[str, Set[str]]):
        self.name: str = name
        self.description: str = description
        self.output: FormulationDescriptor = output
        self.id_to_quantity: Mapping[str, RealDescriptor] = id_to_quantity
        self.labels: Mapping[str, Set[str]] = labels

    def __str__(self):
        return '<IngredientsToFormulationPredictor {!r}>'.format(self.name)



from deprecation import deprecated
from typing import List, TypeVar

from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties


Self = TypeVar('Self', bound='Resource')


# This class is the base type for the new module endpoints. Currently, it only covers predictors,
# but should be applicable for others as they're written.
class EngineResource(Resource[Self]):
    """Base resource for predictors, including metadata representation."""

    created_by = properties.Optional(properties.UUID, 'metadata.created.user', serializable=False)
    """:Optional[UUID]: id of the user who created the resource"""
    create_time = properties.Optional(properties.Datetime, 'metadata.created.time',
                                      serializable=False)
    """:Optional[datetime]: date and time at which the resource was created"""

    updated_by = properties.Optional(properties.UUID, 'metadata.updated.user',
                                     serializable=False)
    """:Optional[UUID]: id of the user who most recently updated the resource,
    if it has been updated"""
    update_time = properties.Optional(properties.Datetime, 'metadata.updated.time',
                                      serializable=False)
    """:Optional[datetime]: date and time at which the resource was most recently updated,
    if it has been updated"""

    archived_by = properties.Optional(properties.UUID, 'metadata.archived.user',
                                      serializable=False)
    """:Optional[UUID]: id of the user who archived the resource, if it has been archived"""
    archive_time = properties.Optional(properties.Datetime, 'metadata.archived.time',
                                       serializable=False)
    """:Optional[datetime]: date and time at which the resource was archived,
    if it has been archived"""

    status = properties.Optional(properties.String(), 'metadata.status.name', serializable=False)
    """:Optional[str]: short description of the resource's status"""
    status_info = properties.Optional(properties.List(properties.String()), 'metadata.status.info',
                                      serializable=False)

    # Due to the way object construction is done at present, __init__ is not executed on Resource
    # objects, so initializing _archived doesn't work.
    _archived = properties.Optional(properties.Boolean(), '', default=None, serializable=False,
                                    deserializable=False)

    _resource_type = ResourceTypeEnum.MODULE

    @property
    def is_archived(self):
        """:bool: whether the resource is archived (hidden but not deleted)."""
        return self.archived_by is not None

    @property
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0",
                details="Please use the 'is_archived' property instead.'")
    def archived(self):
        """[DEPRECATED] whether the resource is archived."""
        return self.is_archived

    @archived.setter
    @deprecated(deprecated_in="1.31.0", removed_in="2.0.0",
                details="Please use archive() and restore() on PredictorCollection instead.")
    def archived(self, value):
        self._archived = value

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental(self) -> bool:
        """[DEPRECATED] whether the execution is experimental (newer, less well-tested)."""  # noqa - insisting this docstring is a signature
        return False

    @property
    @deprecated(deprecated_in="1.25.0", removed_in="2.0.0")
    def experimental_reasons(self) -> List[str]:
        """[DEPRECATED] human-readable reasons why the execution is experimental."""
        return []

    def _post_dump(self, data: dict) -> dict:
        # Only the data portion of an entity is sent to the server.
        data = data["data"]

        # Currently, name and description exists on both the data envelope and the config.
        data["instance"]["name"] = data["name"]
        data["instance"]["description"] = data["description"]
        return data



from setuptools import setup, find_packages
from os import path

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

about = {}
with open(path.join(this_directory, 'src', 'citrine', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setup(name='citrine',
      # Update this in src/citrine/__version__.py
      version=about['__version__'],
      url='http://github.com/CitrineInformatics/citrine-python',
      description='Python library for the Citrine Platform',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Citrine Informatics',
      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      install_requires=[
          "requests>=2.25.1,<3",
          "pyjwt>=1.7.1,<2",
          "arrow>=0.15.4,<0.16",
          "gemd>=1.9.0,<2",
          "boto3>=1.17.93,<2",
          "botocore>=1.20.95,<2",
          "deprecation>=2.0.7,<3",
          "urllib3>=1.26.5,<2",
          "tqdm>=4.62.3,<5"
      ],
      extras_require={
          "builders": [
              "pandas>=1.1.5,<2"
          ],
          "../tests": [
              "factory-boy>=2.12.0,<3"
              "mock>=3.0.5,<4",
              "pandas>=1.1.5,<2",
              "pytest>=6.2.5,<7",
              "pytz>=2019.3"
              "requests-mock>=1.7.0,<2",
          ]
      },
      classifiers=[
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
      ],
)



"""Tools for working with Descriptors."""
from typing import Type, Set

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties

__all__ = ['Descriptor',
           'RealDescriptor',
           'IntegerDescriptor',
           'ChemicalFormulaDescriptor',
           'MolecularStructureDescriptor',
           'CategoricalDescriptor',
           'FormulationDescriptor']


class Descriptor(PolymorphicSerializable['Descriptor']):
    """A Descriptor describes the range of values that a quantity can take on.

    Abstract type that returns the proper type given a serialized dict.
    """

    key = properties.String('descriptor_key')

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        return {
            "Categorical": CategoricalDescriptor,
            "Formulation": FormulationDescriptor,
            "Inorganic": ChemicalFormulaDescriptor,
            "Organic": MolecularStructureDescriptor,
            "Real": RealDescriptor,
            "Integer": IntegerDescriptor,
        }[data["type"]]

    def _equals(self, other, attrs):
        """Check to see if the attrs from the other instance match this instance.

        Returns True if all of the attribute names from attrs match between this and
        the other instance, else False. A missing attribute from this instance will
        raise an AttributeError, and False if missing from the other instance.

        Parameters
        ----------
        other: Description
            the Description instance to compare to
        attrs: List[str]
            A list of attribute names to lookup and compare

        """
        # Check that all attrs exist on this object and raise an AttributeError if not.
        [self.__getattribute__(key) for key in attrs]

        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False


class RealDescriptor(Serializable['RealDescriptor'], Descriptor):
    """A descriptor to hold real-valued numbers.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor
    lower_bound: float
        inclusive lower bound for valid real values
    upper_bound: float
        inclusive upper bound for valid real values
    units: str
        units string; use an empty string to denote a dimensionless quantity

    """

    lower_bound = properties.Float('lower_bound')
    upper_bound = properties.Float('upper_bound')
    units = properties.String('units', default='')
    typ = properties.String('type', default='Real', deserializable=False)

    def __eq__(self, other):
        return self._equals(other, ["key", "lower_bound", "upper_bound", "units", "typ"])

    def __init__(self,
                 key: str,
                 *,
                 lower_bound: float,
                 upper_bound: float,
                 units: str):
        self.key: str = key
        self.lower_bound: float = lower_bound
        self.upper_bound: float = upper_bound
        self.units = units

    def __str__(self):
        return "<RealDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "RealDescriptor({}, {}, {}, {})".format(
            self.key, self.lower_bound, self.upper_bound, self.units)


class IntegerDescriptor(Serializable['IntegerDescriptor'], Descriptor):
    """A descriptor to hold integer-valued numbers.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor
    lower_bound: int
        inclusive lower bound for valid integer values
    upper_bound: int
        inclusive upper bound for valid int values

    """

    lower_bound = properties.Integer('lower_bound')
    upper_bound = properties.Integer('upper_bound')
    units = properties.String('units', default='dimensionless')
    typ = properties.String('type', default='Integer', deserializable=False)

    def __eq__(self, other):
        return self._equals(other, ["key", "lower_bound", "upper_bound", "units", "typ"])

    def __init__(self,
                 key: str,
                 *,
                 lower_bound: int,
                 upper_bound: int):
        self.key: str = key
        self.lower_bound: int = lower_bound
        self.upper_bound: int = upper_bound

    def __str__(self):
        return "<IntegerDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "IntegerDescriptor({}, {}, {}, {})".format(
            self.key, self.lower_bound, self.upper_bound, self.units)


class ChemicalFormulaDescriptor(Serializable['ChemicalFormulaDescriptor'], Descriptor):
    """Captures domain-specific context about a stoichiometric chemical formula.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor

    """

    typ = properties.String('type', default='Inorganic', deserializable=False)

    def __eq__(self, other):
        return self._equals(other, ["key", "typ"])

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<ChemicalFormulaDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "ChemicalFormulaDescriptor(key={})".format(self.key)


class MolecularStructureDescriptor(Serializable['MolecularStructureDescriptor'], Descriptor):
    """
    Material descriptor for an organic molecule.

    Accepts SMILES, IUPAC, and InChI String values.

    Parameters
    ----------
    key: str
        The column header key corresponding to this descriptor

    """

    typ = properties.String('type', default='Organic', deserializable=False)

    def __eq__(self, other):
        return self._equals(other, ["key", "typ"])

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<MolecularStructureDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "MolecularStructureDescriptor(key={})".format(self.key)


class CategoricalDescriptor(Serializable['CategoricalDescriptor'], Descriptor):
    """A descriptor to hold categorical variables.

    An exhaustive list of categorical values may be supplied.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor
    categories: Set[str]
        possible categories for this descriptor

    """

    typ = properties.String('type', default='Categorical', deserializable=False)
    categories = properties.Set(properties.String, 'descriptor_values')

    def __eq__(self, other):
        return self._equals(other, ["key", "categories", "typ"])

    def __init__(self, key: str, *, categories: Set[str]):
        self.key: str = key
        for category in categories:
            if not isinstance(category, str):
                raise TypeError("All categories must be strings")
        self.categories: Set[str] = categories

    def __str__(self):
        return "<CategoricalDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "CategoricalDescriptor(key={}, categories={})".format(self.key, self.categories)


class FormulationDescriptor(Serializable['FormulationDescriptor'], Descriptor):
    """A descriptor to hold formulations.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor

    """

    typ = properties.String('type', default='Formulation', deserializable=False)

    def __eq__(self, other):
        return self._equals(other, ["key", "typ"])

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<FormulationDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "FormulationDescriptor(key={})".format(self.key)



from typing import List

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints.constraint import Constraint

__all__ = ['AcceptableCategoriesConstraint']


class AcceptableCategoriesConstraint(Serializable['AcceptableCategoriesConstraint'], Constraint):
    """
    A constraint on a categorical material attribute to be in a set of acceptable values.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to the associated Categorical descriptor
    acceptable_categories: list[str]
        the names of the acceptable categories to constrain to

    """

    descriptor_key = properties.String('descriptor_key')
    acceptable_categories = properties.List(properties.String(), 'acceptable_classes')
    typ = properties.String('type', default='AcceptableCategoriesConstraint')

    def __init__(self,
                 *,
                 descriptor_key: str,
                 acceptable_categories: List[str]):
        self.descriptor_key = descriptor_key
        self.acceptable_categories = acceptable_categories

    def __str__(self):
        return '<AcceptableCategoriesConstraint {!r}>'.format(self.descriptor_key)



"""Resources that represent measurement run data objects."""
from typing import List, Dict, Optional, Type, Iterator, Union
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, Object, Mapping, LinkOrElse
from citrine.resources.data_concepts import DataConcepts, _make_link_by_uid
from citrine.resources.object_runs import ObjectRun, ObjectRunCollection
from gemd.entity.attribute.condition import Condition
from gemd.entity.attribute.parameter import Parameter
from gemd.entity.attribute.property import Property
from gemd.entity.file_link import FileLink
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.object.material_run import MaterialRun as GEMDMaterialRun
from gemd.entity.object.measurement_run import MeasurementRun as GEMDMeasurementRun
from gemd.entity.object.measurement_spec import MeasurementSpec as GEMDMeasurementSpec
from gemd.entity.source.performed_source import PerformedSource


class MeasurementRun(ObjectRun, Resource['MeasurementRun'], GEMDMeasurementRun):
    """
    A measurement run.

    Parameters
    ----------
    name: str
        Name of the measurement run.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the measurement run.
    conditions: List[Condition], optional
        Conditions under which this measurement run occurs.
    parameters: List[Parameter], optional
        Parameters of this measurement run.
    properties: List[Property], optional
        Properties that are measured during this measurement run.
    spec: MeasurementSpec
        The measurement specification of which this is an instance.
    material: MaterialRun
        The material run being measured.
    spec: MaterialSpec
        The material specification of which this is an instance.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.
    source: PerformedSource, optional
        Information about the person who performed the run and when.

    """

    _response_key = GEMDMeasurementRun.typ  # 'measurement_run'

    name = String('name', override=True)
    uids = Mapping(String('scope'), String('id'), 'uids', override=True)
    tags = PropertyOptional(PropertyList(String()), 'tags', override=True)
    notes = PropertyOptional(String(), 'notes', override=True)
    conditions = PropertyOptional(PropertyList(Object(Condition)), 'conditions', override=True)
    parameters = PropertyOptional(PropertyList(Object(Parameter)), 'parameters', override=True)
    properties = PropertyOptional(PropertyList(Object(Property)), 'properties', override=True)
    spec = PropertyOptional(LinkOrElse(), 'spec', override=True)
    material = PropertyOptional(LinkOrElse(), "material", override=True)
    file_links = PropertyOptional(PropertyList(Object(FileLink)), 'file_links', override=True)
    source = PropertyOptional(Object(PerformedSource), "source", override=True)
    typ = String('type')

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 tags: Optional[List[str]] = None,
                 notes: Optional[str] = None,
                 conditions: Optional[List[Condition]] = None,
                 properties: Optional[List[Property]] = None,
                 parameters: Optional[List[Parameter]] = None,
                 spec: Optional[GEMDMeasurementSpec] = None,
                 material: Optional[GEMDMaterialRun] = None,
                 file_links: Optional[List[FileLink]] = None,
                 source: Optional[PerformedSource] = None):
        if uids is None:
            uids = dict()
        DataConcepts.__init__(self, GEMDMeasurementRun.typ)
        GEMDMeasurementRun.__init__(self, name=name, uids=uids,
                                    material=material,
                                    tags=tags, conditions=conditions, properties=properties,
                                    parameters=parameters, spec=spec,
                                    file_links=file_links, notes=notes, source=source)

    def __str__(self):
        return '<Measurement run {!r}>'.format(self.name)


class MeasurementRunCollection(ObjectRunCollection[MeasurementRun]):
    """Represents the collection of all measurement runs associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/measurement-runs'
    _dataset_agnostic_path_template = 'projects/{project_id}/measurement-runs'
    _individual_key = 'measurement_run'
    _collection_key = 'measurement_runs'
    _resource = MeasurementRun

    @classmethod
    def get_type(cls) -> Type[MeasurementRun]:
        """Return the resource type in the collection."""
        return MeasurementRun

    def list_by_spec(self,
                     uid: Union[UUID, str, LinkByUID, GEMDMeasurementSpec], *,
                     scope: Optional[str] = None) -> Iterator[MeasurementRun]:
        """
        [ALPHA] Get the measurement runs using the specified measurement spec.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDMeasurementSpec]
            A representation of the measurement spec whose measurement run usages are to be located
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid, defaults to Citrine scope ("id")

        Returns
        -------
        Iterator[MeasurementRun]
            The measurement runs using the specified measurement spec.

        """
        link = _make_link_by_uid(uid, scope)
        return self._get_relation('measurement-specs', uid=link)

    def list_by_material(self,
                         uid: Union[UUID, str, LinkByUID, GEMDMaterialRun], *,
                         scope: Optional[str] = None) -> Iterator[MeasurementRun]:
        """
        [ALPHA] Get measurements of the specified material.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDMaterialRun]
            A representation of the material whose measurements are to be queried.
        scope: Optional[str]
            [DEPRECATED] use a LinkByUID to specify a custom scope
            The scope of the uid, defaults to Citrine scope ("id")

        Returns
        -------
        Iterator[MeasurementRun]
            The measurements of the specified material

        """
        link = _make_link_by_uid(uid, scope)
        return self._get_relation(relation='material-runs', uid=link)



from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine.informatics.constraints.constraint import Constraint
from citrine.informatics.descriptors import FormulationDescriptor

__all__ = ['LabelFractionConstraint']


class LabelFractionConstraint(Serializable['LabelFractionConstraint'], Constraint):
    """Represents a constraint on the total amount of ingredients with a given label.

    Parameters
    ----------
    formulation_descriptor: FormulationDescriptor
        descriptor to constrain
    label: str
        ingredient label to constrain
    min: float
        minimum value
    max: float
        maximum value
    is_required: bool, optional
        whether this ingredient is required.
        If ``True``, the label must be present and its value must be within the
        specified range. if ``False``, the label must be within the specified range only if
        it's present in the formulation, i.e., the value can be 0 or on the range ``[min, max]``.

    """

    formulation_descriptor = properties.Object(FormulationDescriptor, 'formulation_descriptor')
    label = properties.String('label')
    min = properties.Optional(properties.Float, 'min')
    max = properties.Optional(properties.Float, 'max')
    is_required = properties.Boolean('is_required')
    typ = properties.String('type', default='LabelFractionConstraint')

    def __init__(self, *,
                 formulation_descriptor: FormulationDescriptor,
                 label: str,
                 min: float,
                 max: float,
                 is_required: bool = True):
        self.formulation_descriptor: FormulationDescriptor = formulation_descriptor
        self.label: str = label
        self.min: float = min
        self.max: float = max
        self.is_required: bool = is_required

    def __str__(self):
        return '<LabelFractionConstraint {!r}::{!r}>'.format(
            self.formulation_descriptor.key,
            self.label
        )
